<?php

namespace Slack\Blockkit;

use Slack\BlockKit;
use Slack\Blockkit\Blocks\Divider;
use Slack\Blockkit\Blocks\Elements\Text;
use Slack\Blockkit\Blocks\Section;
use Slack\Exceptions\SlackSyntaxException;

/**
 * This class creates a slack Modal Response
 */
final class Modal extends BlockKit
{
	protected const LIMITS = [
		'callback_id' => 255,
		'close' => 24,
		'private_metadata' => 3000,
		'submit' => 24,
		'text' => 24,
	];

	private const MAX_BLOCKS = 100;

	private $action = NULL;

	public function __construct(Text $title=NULL,string $type='modal')
	{
		parent::__construct();

		if (! in_array($type,['modal','home']))
			throw new SlackSyntaxException(sprintf('Unknown type %s',$type));

		if ($type != 'modal' && $title)
			throw new SlackSyntaxException(sprintf('Titles are not required for %s',$type));

		if ($title) {
			if ($title->type != 'plain_text')
				throw new SlackSyntaxException(sprintf('Text must be plain_text not %s',$title->type));

			if (strlen($title->text) > self::LIMITS['text'])
				throw new SlackSyntaxException(sprintf('Text must be %d chars or less',self::LIMITS['text']));

			$this->title = $title;
		}

		$this->type = $type;

		$this->blocks = collect();
	}

	/**
	 * Add a block to this modal
	 *
	 * @param Blocks $block
	 * @return $this
	 * @throws SlackSyntaxException
	 */
	public function addBlock(Blocks $block): self
	{
		$this->blocks->push($block);

		if ($x=$this->blocks->count() > self::MAX_BLOCKS)
			throw new SlackSyntaxException(sprintf('Modal can only have %d blocks',self::MAX_BLOCKS));

		return $this;
	}

	/**
	 * The data that will be returned when converted to JSON.
	 */
	public function jsonSerialize()
	{
		switch ($this->action) {
			case 'clear':
				return ['response_action'=>'clear'];

			case 'update':
				return ['response_action'=>'update','view'=>parent::jsonSerialize()];

			default:
				return parent::jsonSerialize();
		}
	}

	/* OPTIONAL ITEMS */

	public function action(string $string)
	{
		if (! in_array($string,['clear','update']))
			throw new SlackSyntaxException(sprintf('Unknown action %s',$string));

		$this->action = $string;

		return $this;
	}

	public function clear_on_close(bool $bool): self
	{
		if ($this->type != 'modal')
			throw new SlackSyntaxException(sprintf('clear_on_close is not required for %s',$this->type));

		$this->clear_on_close = $bool;

		return $this;
	}

	public function callback_id(string $string): self
	{
		$this->callback_id = $this->validate('callback_id',$string);

		return $this;
	}

	public function close(Text $text): self
	{
		if ($this->type != 'modal')
			throw new SlackSyntaxException(sprintf('Close is not required for %s',$this->type));

		if ($text->type != 'plain_text')
			throw new SlackSyntaxException(sprintf('Text must be plain_text not %s',$text->type));

		if (strlen($text->text) > self::LIMITS['close'])
			throw new SlackSyntaxException(sprintf('Text must be %d chars or less',self::LIMITS['close']));

		$this->close = $text;

		return $this;
	}

	// This is a helper method
	public function divider(): self
	{
		$this->blocks->push(Divider::item());

		return $this;
	}

	/**
	 * @param string $string
	 * @return $this
	 * @note A custom identifier that must be unique for all views on a per-team basis.
	 */
	public function external_id(string $string): self
	{
		$this->external_id = $string;

		return $this;
	}

	public function notify_on_close(bool $bool): self
	{
		if ($this->type != 'modal')
			throw new SlackSyntaxException(sprintf('notify_on_close is not required for %s',$this->type));

		$this->notify_on_close = $bool;

		return $this;
	}

	public function private_metadata(string $string): self
	{
		$this->private_metadata = $this->validate('private_metadata',$string);

		return $this;
	}

	// This is a helper method
	public function spacer(): self
	{
		$this->blocks->push(Section::item(Text::item(' ')));

		return $this;
	}

	public function submit(Text $text): self
	{
		if ($this->type != 'modal')
			throw new SlackSyntaxException(sprintf('Submit is not required for %s',$this->type));

		if ($text->type != 'plain_text')
			throw new SlackSyntaxException(sprintf('Text must be plain_text not %s',$text->type));

		if (strlen($text->text) > self::LIMITS['submit'])
			throw new SlackSyntaxException(sprintf('Text must be %d chars or less',self::LIMITS['submit']));

		$this->submit = $text;

		return $this;
	}

	public function submit_disabled(bool $bool): self
	{
		if ($this->type != 'modal')
			throw new SlackSyntaxException(sprintf('submit_disabled is not required for %s',$this->type));

		$this->submit_disabled = $bool;

		return $this;
	}
}