<?php

namespace Slack\Interactive;

use Illuminate\Support\Arr;
use Illuminate\Support\Collection;

use Slack\Models\Channel;

/**
 * This class handles when the users responds with a Block Action event
 *
 * EG:
 * [type] => block_actions
 * [user] => stdClass Object
 * (
 *   [id] => {SLACKUSER}
 *   [username] => {SLACKUSER}
 *   [name] => {SLACKUSER}
 *   [team_id] => {SLACKTEAM}
 * )
 * [api_app_id] => A4TCZ007N
 * [token] => {SLACKTOKEN}
 * [container] => stdClass Object
 * (
 *   [type] => view
 *   [view_id] => V018HRRS38R
 * )
 * [trigger_id] => 1346041864311.39333245939.7c6adb7bca538143098386f07effa532
 * [team] => stdClass Object
 * (
 *   [id] => {SLACKTEAM}
 *   [domain] => {SLACKDOMAIN}
 * )
 * [view] => stdClass Object
 * (
 *   ...
 * )
 * [actions] => Array
 * (
 *   [0] => stdClass Object
 *   (
 *     [action_id] => faq_product
 *     [block_id] => IRFcN
 *     [text] => stdClass Object
 *     (
 *       [type] => plain_text
 *       [text] => ASK QUESTION
 *       [emoji] => 1
 *     )
 *     [value] => question_new
 *     [type] => button
 *     [action_ts] => 1600065294.860855
 *   )
 * )
 */
final class BlockActions extends Base
{
	private const LOGKEY = 'IBA';

	public function __get($key)
	{
		switch ($key) {
			case 'action':
				return Arr::get(object_get($this->_data,'actions'),$this->index);

			// An event can have more than 1 action, each action can have 1 value.
			case 'action_key':
				return $this->keyitem('id',$this->action_id);

			case 'action_value':
				return $this->keyitem('value',$this->action_id);

			// Interactive Messages have a block_id.
			case 'block_key':
				return $this->keyitem('id',$this->block_id);

			case 'block_value':
				return $this->keyitem('value',$this->block_id);

			case 'actions':
			case 'response_url':
				return object_get($this->_data,$key);

			case 'action_id':
			case 'block_id':
				return object_get($this->action,$key);

			case 'callback_id':
				switch ($this->container_type) {
					case 'view':
						// For app hometab, the callback is in the view->callback_id array.
						return object_get($this->_data,sprintf('%s.%s',$this->container_type,$key));

					case 'message':
					case 'message_attachment':
						return NULL;

					default:
						throw new \Exception('Unknown container type: '.$this->container_type);
				}

			case 'container_type':
				return object_get($this->_data,'container.type');

			case 'channel_id':
				return object_get($this->_data,'channel.id') ?: Channel::findOrFail($this->keyitem('value',object_get($this->action,'action_id')))->channel_id;

			case 'keys':
				return collect(object_get($this->_data,'view.blocks'))->pluck('accessory.action_id');

			// For Block Actions that are messages
			case 'message_ts':
				switch ($this->container_type) {
					// Ephemeral messages
					case 'message':
						return object_get($this->_data,'container.message_ts');

					default:
						return object_get($this->_data,'message.ts');
				}

			case 'team_id':                                 // view.team_id represent workspace publishing view
				return object_get($this->_data,'user.team_id');

			case 'view_id':
				return object_get($this->_data,'view.id');

			case 'value':
				switch (object_get($this->action,'type')) {
					case 'external_select':
					case 'overflow':
					case 'static_select':
						return object_get($this->action,'selected_option.value');
					case 'multi_static_select':
						return object_get($this->action,'selected_options.value');

					case 'button':
					default:
						return object_get($this->action,$key);
				}

			case 'value_count':
				switch ($x=object_get($this->action,'type')) {
					// @todo To Check
					case 'external_select':
					// @todo To Check
					case 'overflow':
					// @todo To Check
						throw new \Exception('To be implemented: ',$x);
					case 'static_select':
						return count(object_get($this->action,'selected_option'));
					case 'multi_static_select':
						return collect(object_get($this->action,'selected_options'))->pluck('value');
					default:
						return count(object_get($this->action,'value'));
				}

			default:
				return parent::__get($key);
		}
	}

	/**
	 * Some block actions are triggered by messages, and thus dont have a callback_id
	 *
	 * @return bool
	 */
	public function isMessage(): bool
	{
		return (bool)object_get($this->_data,'message');
	}

	/**
	 * Get the selected options from a block action actions array
	 *
	 * @return Collection
	 */
	public function selected_options(): Collection
	{
		$result = collect();

		foreach (Arr::get(object_get($this->_data,'actions'),'0')->selected_options as $option) {
			$result->push($option->value);
		}

		return $result;
	}
}