<?php

namespace Slack\Message;

use Slack\BlockKit;
use Slack\Blockkit\Blocks\Elements\Text;
use Slack\Exceptions\SlackSyntaxException;

/**
 * Class MessageAttachmentAction - Slack Message Attachments Actions
 * Represents a Single Action for a Slack Message Attachment
 *
 * @package App\Slack\Message
 * @note These are now legacy - use blocks instead
 * @note more information on this available https://api.slack.com/legacy/interactive-message-field-guide
 */
final class AttachmentAction extends BlockKit
{
	protected const LIMITS = [
		'text' => 30,
		'value' => 2000,
	];

	private const DATA_SOURCES = ['static','users','channels','conversastions','external'];
	private const STYLES = ['default','danger','primary'];
	private const TYPES = ['button','select'];

	// @todo options
	// @todo option_groups
	// @todo selected_options

	public function __construct(string $text,string $name,string $type)
	{
		parent::__construct();

		if (! in_array($type,self::TYPES))
			throw new SlackSyntaxException(sprintf('Type [%s] not valid',$type));

		$this->type = $type;
		switch ($type) {
			case 'button':
				$this->action_id = $name;
				$this->text = Text::item($this->validate('text',$text),'plain_text');
				break;

			case 'select':
				$this->name = $name;
				$this->text = $this->validate('text',$text);
				break;
		}
	}

	public static function item(string $text,string $name,string $type): self
	{
		return new self($text,$name,$type);
	}

	/* OPTIONAL ITEMS */

	/**
	 * Set a confirmation diaglog when this action is selected
	 *
	 * @param string $title
	 * @param string $text
	 * @param string $ok_text
	 * @param string $dismiss_text
	 * @return $this
	 */
	public function confirm(string $title,string $text,string $ok_text,string $dismiss_text): self
	{
		$this->confirm = [
			'title' => $title,
			'text' => $text,
			'ok_text' => $ok_text,
			'dismiss_text' => $dismiss_text
		];

		return $this;
	}

	public function data_source(string $string): self
	{
		if (! in_array($string,self::DATA_SOURCES))
			throw new SlackSyntaxException(sprintf('Type [%s] not valid',$string));

		$this->data_source = $string;

		return $this;
	}

	public function min_query_length(int $int): self
	{
		$this->min_query_length = $int;

		return $this;
	}

	/**
	 * Set the text displayed in the action
	 *
	 * @param string $string
	 * @return $this
	 * @throws SlackSyntaxException
	 */
	public function style(string $string): self
	{
		if (! in_array($string,self::STYLES))
			throw new SlackSyntaxException(sprintf('Type [%s] not valid',$string));

		$this->style = $string;

		return $this;
	}

	/**
	 * Set the value for the action
	 *
	 * @param string $string
	 * @return $this
	 * @throws SlackSyntaxException
	 */
	public function value(string $string): self
	{
		$this->value = $this->validate('value',$string);

		return $this;
	}
}